# Copyright (c) 2015  aggftw@gmail.com
# Distributed under the terms of the Modified BSD License.
# from sparkmagic package
NO_AUTH = "None"
AUTH_KERBEROS = "Kerberos"
AUTH_BASIC = "Basic_Access"
AUTHS_SUPPORTED = [NO_AUTH, AUTH_KERBEROS, AUTH_BASIC]


CONFIGURABLE_RETRY = "configurable"
LINEAR_RETRY = "linear"
LINEAR_RETRY_ANY_ERROR = "linear_retry_any_error"
